<?php
/**
 * @file
 * Token callbacks for the AET module.
 */

/**
 * Implements hook_token_info().
 *
 * Declaring the AET tokens.
 */
function aet_token_info() {

  // Initializing the $info array.
  $info = array('types' => array(), 'tokens' => array());

  $info['types']['aet'] = array();
  $info['types']['aet']['name'] = t('Advanced Entity Tokens');
  $info['types']['aet']['description'] = t('AET Description');

  // Getting the information arrays on all the entities.
  $entities = \Drupal::service('entity_type.manager')->getDefinitions();

  // Looping the entity information array to add the token types.
  foreach ($entities as $entity_type => $entity_info) {

    /** @var \Drupal\Core\Entity\EntityType $entity_info */
    $token_type = $entity_info->get('additional')['token_type'] ?: $entity_type;

    if (empty($info['types'])) {
      $info['types'] = array();
    }

    if (empty($info['tokens'])) {
      $info['tokens'] = array();
    }

    $entity_id_key = $entity_info->getKey('id');

    // Adding the entity element token.
    // The token key consists of the entity key id and a question mark
    // to mark this as the place to enter the entity's id.
    $info['tokens']['aet'][$token_type] = array();

    $info['tokens']['aet'][$token_type]['name']
    = 'AET ' . $entity_info->getBundleLabel() . ' entity';

    $info['tokens']['aet'][$token_type]['description']
    = t("Reference a @entity_name entity by using it's @id_key",
      array(
        '@entity_name' => $entity_info->getBundleLabel(),
        '@id_key' => $entity_id_key
      )
    );

    $info['tokens']['aet'][$token_type]['dynamic']
    = TRUE;

    $info['tokens']['aet'][$token_type]['type']
    = $token_type;

  }

  return $info;
}

/**
 * Implements hook_tokens().
 */
function aet_tokens($type, $tokens, array $data, array $options, \Drupal\Core\Render\BubbleableMetadata $bubbleable_metadata) {
  // Initiating the $replacements array.
  $replacements = array();
  // TODO Add support $options['callback']
  // TODO Add support $options['sanitize']
  // TODO Add support $options['language']
  $options += _aet_default_options();
  // Only tokens from the aet type are supported.
  if ($type === 'aet') {
    // Gets the information array on all the available entities in this drupal
    // site.
    $entities = \Drupal::service('entity_type.manager')->getDefinitions();

    // Loops the $entities array to check if they exist in the $tokens array.
    foreach ($entities as $entity_type => $entity_info) {

      /** @var \Drupal\Core\Entity\EntityType $entity_info */
      $token_type = $entity_info->get('additional')['token_type'] ?: $entity_type;

      $data += array('entity_type' => $entity_type);

      // The aet_entity_tokens function replaces the next level of the aet
      // tokens.
      // E.G. [aet:node:1], [aet:node:label]
      $replacements += _aet_entity_tokens($token_type,
        \Drupal::service('token')->findWithPrefix($tokens, $token_type),
        $data, $options, $bubbleable_metadata);
    }
    // Calls aet_clear_tokens to clear unreplaced tokens if the option is
    // flagged.
    aet_clear_tokens($tokens, $replacements, $options['clear']);
  }
  // Returns the $replacements array.
  return $replacements;
}

/**
 * INTERNAL implementation of hook_tokens specified for entities.
 *
 * This function implements the entity level of AET.
 * E.G. [aet:node:1] [aet:file:label]
 */
function _aet_entity_tokens($type, $tokens, array $data = array(), array $options = array(), \Drupal\Core\Render\BubbleableMetadata $bubbleable_metadata) {

  // Initiating the $replacements array.
  $replacements = array();

  // TODO Add support $options['language']
  // TODO Add support $options['callback']
  // TODO Add support $options['sanitize']
  $options += _aet_default_options();

  $entity_type_manager = \Drupal::service('entity_type.manager');
  $entities = $entity_type_manager->getDefinitions();

  $entity_type_id = '';

  $entity_type_info = NULL;

  foreach ($entities as $key => $value) {

    // Gets the information array on all the available entities in this drupal
    // site.
    /** @var \Drupal\Core\Entity\EntityType $value */
    $token_type = $value->get('additional')['token_type'] ?: NULL;

    if ($token_type === $type) {
      $entity_type_id = $key;
      $entity_type_info = $value;
    }
  }

  if (!empty($entity_type_info)) {
    foreach ($tokens as $key => $original) {

      // The position of the first colon is used for both checking if a secound
      // part to the key exists and to seperate between it (the first part) and
      // the rest of the key.
      $first_colon_pos = strpos($key, ':');

      // The $first_key is the part of the key until the first colon if exist,
      // or the full $key if no colon exists.
      $first_key = $first_colon_pos ? substr($key, 0, $first_colon_pos) : $key;

      // Checks if the $first_key is numeric - the entity id.
      if (_aet_is_valid($first_key)) {

        // Renames the $first_key as $entity_id for better readability.
        $entity_id = $first_key;

        $id_key = $entity_type_info->getKey('id');

        // Check if the entity to load is currently being viewed or if the
        // the rendered token origins from itself.
        if (aet_entity_is_page($entity_id, $id_key)
          || aet_entity_is_viewed($entity_type_id, $id_key, $entity_id)) {

          // To avoid recursion we skip this entity.
          drupal_set_message(t('Recursive token use detected.'), 'warning');
          continue;
        }

        // Loads the entity. entity_loads loads the entity into an assosative
        // array with the $entity_id as the array key. we use array_values to
        // move the entities array into a simple array.
        if ($entity = $entity_type_manager->getStorage($entity_type_id)->load($entity_id)) {

          $user = \Drupal::currentUser();

          // Checks if the user has access to view this entity.
          if (!$entity->access('view', $user)) {
            continue;
          }

          // Checks if the $tokens array contain the $entity_id as one of the
          // "to be replaced" tokens.
          if (isset($tokens[$entity_id])) {

            // The $view_mode of the entity will default to token if exists or
            // full if it doesn't. use another chaining level to specify the
            // display.
            $view_modes = \Drupal::service('entity_display.repository')->getViewModes($entity_type_id);
            $view_mode = isset($view_modes['token']) ? 'token' : 'full';

            // Get the renderable array of the entity with entity_view.
            $renderable_entity
            = $entity_type_manager->getViewBuilder($entity_type_id)->view($entity, $view_mode);

            // Render the entity.
            $rendered_entity = render($renderable_entity);

            // Add the rendered entity to the replacements array.
            $replacements[$tokens[$entity_id]] = $rendered_entity;
          }

          $token_service = \Drupal::service('token');
          $replacements += _aet_entity_view_tokens($type,
            $token_service->findWithPrefix($tokens, $entity_id),
            array($type => $entity), $options, $bubbleable_metadata);

          // I'm calling the token_generate to the rest of the chaining at this
          // point - after the entity type and nid have been figured out.
          $replacements += $token_service->generate($type,
            $token_service->findWithPrefix($tokens, $entity_id),
            array($type => $entity), $options, $bubbleable_metadata);
        }
      }
    }

    // Calls aet_clear_tokens to clear unreplaced tokens if the option is
    // flagged.
    aet_clear_tokens($tokens, $replacements, $options['clear']);
  }

  // Returns the $replacements array.
  return $replacements;
}


/**
 * This is the main implementation of hook_tokens.
 *
 * The main purpose of this hook implementation is to identify the entity type
 * and call aet_entity_token on it.
 */
function _aet_entity_view_tokens($type, $tokens, array $data = array(), array $options = array(), \Drupal\Core\Render\BubbleableMetadata $bubbleable_metadata) {

  // Initiating the $replacements array.
  $replacements = array();

  // TODO Add support $options['language']
  // TODO Add support $options['callback']
  // TODO Add support $options['sanitize']
  $options += _aet_default_options();

  // Gets the information array on all the available entities in this drupal
  // site.
  $entity_type_manager = \Drupal::service('entity_type.manager');
  $entities = $entity_type_manager->getDefinitions();

  $entity_type = '';

  $entity_info = NULL;

  foreach ($entities as $key => $value) {

    /** @var \Drupal\Core\Entity\EntityType $value */
    $token_type = $value->get('additional')['token_type'] ?: NULL;

    if ($token_type === $type) {
      $entity_type = $key;
      $entity_info = $value;
    }
  }
  // Checks if $entity_info is not empty.
  if (!empty($entity_info) && !empty($data[$type])) {

    $entity = $data[$type];

    /** @var \Drupal\Core\Entity\EntityType $entity_info */
    foreach ([] as $key => $value) {

      if (!empty($tokens['view-' . $key])) {
        // Renames $key to $view_mode for better readability.
        $view_mode = $key;

        // Get the renderable array of the entity with entity_view.
        $renderable_entity
        = $entity_type_manager->getViewBuilder($entity_type)->view($entity, $view_mode);

        // Render the entity.
        $rendered_entity = render($renderable_entity);

        // Add the rendered entity to the replacements array.
        $replacements[$tokens['view-' . $key]] = $rendered_entity;
      }
    }
  }

  // Returns the $replacements array.
  return $replacements;
}

function _aet_default_options() {
  global $language;
  return array(
    'callback' => NULL,
    'sanitize' => FALSE,
    'language' => $language,
    'clear' => FALSE,
  );
}
